#ifndef CAMERA_HPP
#define CAMERA_HPP

#include <glm/vec3.hpp>
#include <glm/mat4x4.hpp>
#include "sceneObject.hpp"

struct GLFWwindow;

//! Implements the camera movement with user input.
class Camera : public SceneObject {
public:
    //! Construct new camera.
    /*!
     * \details The camera is a SceneObject, but it has also a view matrix.
     * The computation of the view matrix is done from the horizontal and vertical angles of the camera.
     * From these Spherical coordinates the Cartesian coordinates are computed,
     * these are the direction and the right vector.
     * The cross product of this two vector gives the up vector.
     * Then using the glm::lookAt function, the view matrix is computed.
     * The camera has a projection matrix.
     * On zooming this projection matrix is recomputed using the glm::perspective function.
     * When an object is being rendered, the view matrix of the camera is returned.
     * \sa http://www.opengl-tutorial.org/beginners-tutorials/tutorial-6-keyboard-and-mouse
     * \todo fix documentation
     */
    Camera(const std::string& name,
           Scene* scene,
           const glm::vec3& eye = glm::vec3(0, -10, 5),
           float vAngle = -3.14f,
           float hAngle = 3.14f,
           float fov = 60.0f,
           float near = 0.01f,
           float far = 20.0f);

    //! Construct new camera.
    /*!
     *
     */
    Camera(const std::string& name,
           Scene* scene,
           glm::vec3& pos,
           glm::vec3& up,
           glm::vec3& eye,
           float fov,
           float near,
           float far);

    virtual ~Camera();

    void draw() const;
    bool animate(double time);
    void update(double deltaT);
    void setShader(Shader*) {}

	bool movement();

    float getNearClip() { return nearClip; }
    float getFarClip() { return farClip; } 

	void initializeLookupVector();
    void setViewMatrix(const glm::mat4& modelMatrix);
    glm::mat4 getViewMatrix() const { return viewMatrix; }
    glm::mat4 getProjectionMatrix() const { return projectionMatrix; }

protected:
    double cursorX;
    double cursorY;

    float fov;
    float farClip;
    float nearClip;
    glm::vec3 up;
    glm::vec3 eye;
    glm::vec3 pos;
    glm::mat4 viewMatrix;
    glm::mat4 projectionMatrix;

private:
	bool moved;
	bool initState;
};

#endif //CAMERA_HPP
