#ifndef SUN_HPP
#define SUN_HPP

#include <vector>
#include <glm/vec3.hpp>
#include <glm/vec4.hpp>
#include <glm/mat4x4.hpp>

#include "light.hpp"

/*! Implements the sun in the scene together with shadow mapping techniques.
* \author Felix
* \details The Sun is a derivation of the Light class which is itself a SceneObject.
* In addition to the methods from SceneObject, a Sun has methods that use different shadow mapping techniques.
* The method drawOmni supports Omnidirectional Shadow mapping where every SceneObject is rendered into a cubemap texture.
* This texture will then be used for the shadow map lookup. In addition, we implemented Shadow volumes and therefore 2 renderpasses are used.
* The method ambientPass and SVPass compute the shadow volumes using each SceneObject in the 2 seperate renderpasses.
* For each seperate renderpass, effectobjects are used that are passed onwards to the individual SceneObjects that will be rendered.
* Framebuffer Objects are used for storing different shadow textures.
* \todo fix documentation
*/
class Sun : public Light
{
public:

	//! Construct new sun.
    Sun(const std::string& name, Scene* scene = 0, Model* model = 0, Shader* shader = 0, const glm::mat4& modelMatrix = glm::mat4(1));
    virtual ~Sun();

    void draw() const;
    void reset();
    void update(double deltaT);
    void setShader(Shader* val);

    void drawOmni() const;
    void setUpAmbientPass();
    void ambientPass();
    void SVPass();

    void setUpColorTexture();
    void setUpDepthTexture();
    GLuint getDepthTexture() const { return depthTexture; }

private:
    class EffectObjectAmbientPass;
    friend class EffectObjectAmbientPass;
    class EffectObjectSVPass;
    friend class EffectObjectSVPass;
    class EffectObjectCube;
    friend class EffectObjectCube;
    class EffectObjectCubeInstanced;
    friend class EffectObjectCubeInstanced;

    GLuint frameBuffer;
    GLuint depthTexture;
    GLuint colorTexture;
    GLuint cubeTexture;

    std::vector<glm::mat4> viewMatrices;
    
    int renderPasses;
    GLenum cubeFaces[6];

    void setViewMatrices();
};

#endif //SUN_HPP
